{{
Purpose of PASMstore is to have an environment in which PASM code can be created,
tested and then stored on an SD card. Of course it can be adapted to provide this
functionality for any kind of mass storage.
These PASM code snippets can then be loaded by other programs. This is good for
dynamic program reload (e.g. drivers) with a small memory footprint. In this way the whole 32k of
HUB RAM can be used for other application code or data.

This program uses the SD card driver available at the ObjectExchange.

Please make sure that you modify the SDCardPinBase to match your SD card connection.
Please note the comments in SDSPIQASM on how to connect the SD card if you don't
have a board including a SD card slot.

I like to see what a program is doing. That's why I used the serial driver to
show some short messages. For bigger libraries of PASM it's useful for debugging
of course.

In a later version all the FullDuplexSerial code and all the SD card code should
be included directly in this file. This way the PASMstore can be run with F10 whereas
the BigProgram can be run with Ctrl + F10.

The content of a *.cog file is simply the length of the PASM code, which is stored in
the first 4 bytes. (Yes, I know that this is by far to long for the propeller, but
the propeller is a 32 bit controller ;o)
Following bytes contain the machine code.
}}
CON
  _clkmode        = xtal1 + pll16x
  _xinfreq        = 5_000_000

  STG_LINE_WIDTH   = 16

  LCDBasePin    = 8

obj
  desc: "COGdesc(Term)"

pub PASMstore | slot
  waitcnt( clkfreq*10 + cnt )
  init( LCDBasePin )
  
  desc.init
  desc.show
  
  desc.setName( string("BenkyLCD") )
  desc.setVersion( $0099 )
  desc.setPSize( 2 )
  desc.setVAdr( $8c00 )
  
  slot := desc.putDescription
  desc.saveDriver( slot, @LCDdrv )
  desc.saveTable

  desc.show
  
  repeat
  
pub init( pinBase )
  basePin := pinBase
  w5ms   := clkfreq/200              ' long  400_000
  w500us := clkfreq/2000             ' long  40_000

DAT
ORG 0
LCDdrv        
              jmp       #LCDdrv_init            ' init-code is only needed once, so I put
                                                ' it to the end. This way the memory at the
                                                ' end can be reused in the following code
initDone
              call      #blnkclr
              mov       cntlp, cnt
        
waitLoop
              'sub   cntlp, cnt                 ' test code to find out the average        
              'neg   cntlp, cntlp               ' runtime
              'wrlong cntlp, spinBuf1
              wrlong    outb, spinBuf           ' done with execution of instruction
waitCmd
              'mov   cntlp, cnt
              test      flags, #1 WC
        if_C  jmp       #doInstr
              call      #doBlink                ' let's see if something has to be done
              call      #doLine                 ' for blinking char or scrolling line
doInstr
              rdlong    cmd, spinBuf WZ         ' next instruction available?
        if_Z  jmp       #waitCmd

              mov       cmdjmp, origjmp         ' calculate the jump-address
              add       cmdjmp, cmd
              rdlong    o1, spinBuf1            ' get the parameter

cmdjmp        jmp       #jmplist-1
origjmp       jmp       #jmplist-1                        
cntlp         long      0

jmplist 
              jmp       #printStr
              jmp       #printNStr
              jmp       #rawCmd
              jmp       #writeData
              jmp       #setChar
              jmp       #setBlink
              jmp       #setRate
              jmp       #shift
              jmp       #setLine
              jmp       #setWidth
              jmp       #sAtomic
              jmp       #eAtomic

sAtomic
'======
              or        flags, #1
              jmp       #waitLoop
eAtomic
'======
              andn      flags, #1
              jmp       #waitLoop
              
' ok, name no longer fits, as the width and a incrementor is set with
' this instrunction.
setWidth
'=======
              mov       line_width, o1
              and       line_width, posfreq_mask
              mov       line_incr, o1
              shr       line_incr, #16
              jmp       #waitLoop

setRate
'======
              mov       blnkfreq, o1
              jmp       #waitLoop

' The upper word tells which of the 4 scrolling lines to use
' and gives an offset to the screen-position for where to start output.
' the lower word is the address of the string in HUB-RAM
'
' xx000000_oooooooo_aaaaaaaaa_aaaaaaaa
' x = Number of scroller to use
' o = offset
' a = address of the text in HUB-RAM
setLine
'======
              mov       t1, o1                            ' extract the upper 16 bit
              mov       movit, movto
              shr       t1, #30 WZ
              and       t1, #3
              shl       t1, #10
              add       movit, t1
              mov       t1, movit
              shr       t1, #9
              movd      movit2, t1
              andn      o1, msbits                        ' clear the index bits

              add       movit2,dst_offset        
movit         mov       line1, o1
movit2        mov       line1, #0                         
              jmp       #waitLoop

movto         mov       line1, o1   

lnval         long      1000000
msbits        long      %11000000_00000000_00000000_00000000

doLine
'-----
              sub       lnval, #1 WZ
        if_NZ jmp       #doLine_ret
              mov       lnval, blnkfreq
              ' shr   lnval, #8

              call      #push_cpos

              mov       in1, i1
              mov       in2, i2
              'mov       in3, i3
              mov       in4, i4
              mov       in5, i5
              mov       t5, #4
                
in1           mov       t3, line1 WZ
              mov       t4, t3
              and       t3, posfreq_mask
              mov       t1, t3
        if_Z  jmp       #doNextLine
              and       t4, ch1_mask
              shr       t4, #16
              mov       o1, #%10000000
              add       o1, t4
              call      #outCmd
in2           add       t1, line1d
              rdbyte    o1, t1 WZ
'in3     if_Z  mov       t1, line1
'        if_Z  and       t1, posfreq_mask
in4     if_Z  mov       line1d, #0
in5           add       line1d, line_incr '#1
              call      #printn

doNextLine
              add       in1, #2
              add       in2, #2
              'add       in3, #2
              add       in4, dst_offset2
              add       in5, dst_offset2
              djnz      t5, #in1

              call      #pop_cpos
        
doLine_ret
              ret
        
i1            mov       t3, line1 WZ
i2            add       t1, line1d
i3      if_Z  mov       t1, line1
i4      if_Z  mov       line1d, #0
i5            add       line1d, line_incr '#1

printn        mov       t2, line_width
printnlp      rdbyte    o1, t1 WZ
        if_Z  mov       t1, t3
        if_Z  rdbyte    o1, t1
              call      #printChar
              add       t1,#1
              djnz      t2,#printnlp
printn_ret
              ret
        
line_width    long      STG_LINE_WIDTH
line_incr     long      1
                
shift
'====
              mov       par1, o1
              and       par1, # $1f
              and       o1,#%10000000 WZ,NR
        
              mov       o1,#%00011010
        if_NZ or        o1,#%00000100         
shiftlp
              call      #outCmd
              sub       par1, #1 WZ
        if_NZ jmp       #shiftlp
                 
              jmp       #waitLoop
        
blnkfreq      long      1000000
blnkval       long      1000000

doBlink
'------
              sub       blnkval, #1 WZ
        if_NZ jmp       #doBlink_ret

              call      #push_cpos
              mov       blnkval, blnkfreq
         
              ' init the loop
              movs      blnklp, #blnktab
              movd      blnkwb, #blnktab 
              mov       t1, #8

              ' loop over blink-table and see if a entry reached due-time
blnklp        mov       t2, blnktab
              and       t2, #$ff WZ, NR 
        if_Z  jmp       #blnknext

              ' if so, set the adress
              mov       o1, t2                        ' extract the adress from the table entry
              and       o1, idxbyte
              shr       o1, #8
              or        o1, #%10000000
              call      #outCmd
        
              ' set blink character1 or 2
              mov       o1, t2
              shr       o1, #16
              and       o1, #$ff
              call      #printChar
              sub       cpos, #1

              ' swap the character bytes for the next loop
              mov       t3, t2
              mov       o1, t2
              and       t2, posfreq_mask

              and       t3, ch1_mask
              shl       t3, #8
              and       o1, ch2_mask
              shr       o1, #8
              or        t2, o1                          ' and copy byte 2 to byte 3 position
              or        t2, t3                        ' and byte 3 to byte 2 position
blnkwb        mov       blnktab, t2                           ' write back the entry        

blnknext
              add       blnklp, #1
              add       blnkwb, dst_offset
              djnz      t1, #blnklp
 
              ' set cursor position back to cpos
              call      #pop_cpos
              'mov   o1, #%10000000
              'or    o1, cpos
              'call  #outCmd
        
doBlink_ret
              ret

dst_offset    long      %01000000000
dst_offset2   long      %10000000000
posfreq_mask  long      $0000ffff
ch1_mask      long      $00ff0000
ch2_mask      long      $ff000000
pos_mask      long      $7f000000

push_cpos
              mov       cpos_bk, cpos
push_cpos_ret
              ret

pop_cpos
              mov   cpos, cpos_bk
              ' set cursor position back to cpos
              'mov   o1, #%10000000
              'or    o1, cpos
              'call  #outCmd
              'mov       t4, o1
              'mov   o1, #%10000000
              'or    o1, cpos
              'call  #outCmd
              'mov       o1, t4
pop_cpos_ret
              ret

set_cpos
              mov       t4, o1
              mov   o1, #%10000000
              or    o1, cpos
              call  #outCmd
              mov       o1, t4
set_cpos_ret
              ret
        
' setBlink expects 4 byte parameters packed into a long
' byte 0 = blink frequency/2 in seconds*10 (from 0,1 to 25,5 seconds)
' byte 1 = index in the blink-table where all blink characters are registered
' byte 2 = character 1, which is shown for byte 1 seconds
' byte 3 = character 2, which is shown for byte 1 seconds
' before setting a blinking character the cursor needs to be set
' to remove a blink, just pass byte 1 and let the rest be zero
setBlink         
'=======
              mov       tmp2, o1                        ' use tmp2 to calculate the offset
              mov       setblnk, cpyblnk
              shr       tmp2, #8
              and       tmp2, #7                          ' limit the index
              shl       tmp2, #9
              add       setblnk, tmp2                     ' add the offset to destination of move
              andn      o1, idxbyte WZ                    ' remove the index byte from parameter
        if_NZ mov       tmp2, cpos                        ' put the cpos in place
        if_NZ shl       tmp2, #8
        if_NZ and       tmp2, idxbyte
        if_NZ or        o1, tmp2                        ' and copy it into the blink table entry
setblnk       mov       blnktab, o1
              jmp       #waitLoop

cpyblnk       mov       blnktab, o1
idxbyte       long      $0000ff00
 
' setChar reads 8 bytes and sends it to the display. Intention is to set
' CGRAM, so that we can display user defined characters. With 5x8 display we
' have 8 user definable characters.
setChar
'======
              mov       par1, o1
              mov       par2,#8
setCharLp
              rdbyte    o1, par1  WZ
              call      #printChar
              add       par1, #1
              sub       par2, #1 WZ
        if_NZ jmp       #setCharLp
              jmp       #waitLoop
        
' rawCmd is a raw interface for the display. Whatever byte you pass will
' be send to the display. This way you can clear the display or set cursor position...
rawCmd
'=====
              and       o1, #%10000000 WZ,NR
        if_NZ mov       cpos, o1
        if_NZ and       cpos, #%01111111 
              call      #outCmd
              jmp       #waitLoop

writeData
'========
              'rdlong o1, spinBuf1
              call      #printChar
              jmp       #waitLoop

        
' printStr outputs a string ( remember, a string ends with a 0 ) at the current position.
' The adress of the string 
printStr
'=======
              ' set cursor position back to cpos
              call      #set_cpos
              'mov       t4, o1
              'mov       o1, #%10000000
              'or        o1, cpos
              'call      #outCmd
              'mov       o1, t4

              mov       par1,o1
printStrLp
              rdbyte    o1, par1  WZ
        if_Z  jmp       #waitLoop

              add       par1, #1
              call      #printChar

              jmp       #printStrLp

printNStr
'========
              ' set cursor position back to cpos
              call      #set_cpos
              'mov       t4, o1
              'mov       o1, #%10000000
              'or        o1, cpos
              'call      #outCmd
              'mov       o1, t4

              mov       par1,o1
              and       par1, posfreq_mask
              shr       t4, #16
printNStrLp
              rdbyte    o1, par1  WZ

              add       par1, #1
              call      #printChar

              djnz      t4, #printNStrLp
              jmp       #waitLoop


' printChar expects the character to be written at the current cursor position in o1.
printChar
'--------
              or        outa, RSpin
              call      #outHigh
              call      #pulse
              call      #outLow
              call      #pulse
              andn      outa, RSpin

              add       cpos, #1

              call      #waitBusy
printChar_ret
              ret

' outLow and outHigh both expect a byte in o1. outLow sets the data pins according to
' the low nibble and outHigh ... well ... guess        
outLow
'-----
              mov       tmp1, o1
              jmp       #out
outHigh
'------
              mov       tmp1, o1                          ' first make a copy of output register
              shr       tmp1, #4                          ' put high nibble to beginning
out           shl       tmp1, dataPin                     ' and shift it to the data pin position
              and       tmp1, DataMask                    ' mask out all other bits
              andn      outa, DataMask                    ' first delete the data bits
              or        outa, tmp1                        ' and then set the ones which are high
outHigh_ret
outLow_ret
              ret

' outCmd awaits the LCD instruction to be passed in o1.
outCmd
'-----
              call      #outHigh
              call      #pulse
              call      #outLow
              call      #pulse

              call      #waitBusy
outCmd_ret
              ret

' pulse expects 3 parameters:
' p1 = pin or mask of pins to set
' p2 = high-time
' p3 = low-time
' It set's the output according to parameter 1 and waits for p2 clock-cycles.
' Then it clears the output pins again and waits for p3 clock-cycles. P3 can be
' set to 0, to ommit waiting.
pulse
'----
              or        outa, p1
              mov       tmp1, cnt
              add       tmp1, p2
              waitcnt   tmp1, p3
              andn      outa, p1
              mov       p3,p3 WZ, NR
        if_NZ waitcnt   tmp1, #0
        
pulse_ret
              ret

' waitBusy waits as long as the display set's the busy flag
waitBusy
'-------
              andn      dira, DataMask                    ' now the data-pins are input
              or        outa, RWpin                       ' set the R/W pin, so we can read

waitAgain
              ' one busy flag retreival needs 2 E-cycles in 4 bit mode        
              mov       tmp1, cnt
              add       tmp1, w500ns
              or        outa, Epin
              waitcnt   tmp1, w500ns
              andn      outa, Epin
              mov       tmp2, ina                         ' read the bits (busy, A6-A4)
              waitcnt   tmp1, w500ns
              and       tmp2, DataMask                    ' extract the data pins
              or        outa, Epin
              waitcnt   tmp1, w500ns
              shr       tmp2, dataPin                     ' and shift the pins to start at bit 0
              andn      outa, Epin
              waitcnt   tmp1, w500ns
              and       tmp2, #%1000 WZ,NR                ' now we can check the busy bit
        if_NZ jmp       #waitAgain                        ' set, then repeat

              andn      outa, RWpin                       ' clear the R/W pin, so we can write
              or        dira, DataMask                    ' and make data pins output
waitBusy_ret
              ret

' here we have the storage area of pin-numbers and mask .... things
' frequently needed in the PASM code        
dataPin              
basePin       long      0
RSpin         long      1
RWpin         long      2
Epin          long      4
DataMask      long      %1111000
BusMask       long      %0000111

w500ns        long      25
w5ms          long      400000
w500us        long      40000

spinBuf       long      0
spinBuf1      long      4
spinBuf2      long      8

' here comes the initialisation of the PASM. As this is only needed to run
' once, all the memory eaten up here can be reused in code above for storage ...
' and is already used, as you can see on the labels ;o)

LCDdrv_init
'----------
              ' first initialize the pointers to the SPIN-communication-buffer
cmd           mov       spinBuf, par
par1          add       spinBuf1, par
par2          add       spinBuf2, par

              ' shift the bitmasks to the basePin
o1            shl       RSpin, basePin
p1            shl       RWpin, basePin
p2            shl       Epin, basePin
p3            shl       DataMask, basePin
tmp1          shl       BusMask, basePin
tmp2          shl       outMask, basePin
t1            add       basePin, #3                        ' after that basePin is reused as dataPin

              ' outa is 0 per default, so the pins can be switched to output
t2            mov       dira, outMask         

              ' now let's initialize the LCD, so it runs in 4-bit-mode
t3            mov       o1, #%00110010
t4            call      #outHigh
t5            mov       p1, Epin
cpos_bk       mov       p2, w500ns
cpos          mov       p3, w5ms
flags         call      #pulse

              ' call outLow
              ' p1-p2 are still set correctly
blnktab       mov       p3, w500us
              call      #pulse

              ' call outLow
              ' p1-p2 are still set correctly
              call      #pulse

              call      #outLow
              call      #pulse

              mov       o1, #%00101100
              call      #outCmd

              ' Display / Cursor / Blink on/off
              mov       o1, #%00001100
              call      #outCmd

              ' Clear
              mov       o1, #%00000001
              call      #outCmd

              ' Mode Increment/Decrement
              mov       o1, #%00000110
              call      #outCmd

              mov       p3, w500ns

              ' Home
line1         mov       o1, #%00000010
line1d        call      #outCmd

line2         jmp       #initDone
line2d
outMask       long      %1111111

line3         long      0
line3d        long      0
line4         long      0
long4d        long      0

blnkclr
'======
              mov       t1, #20
blnkleer
              mov       flags, #0
              add       blnkleer, dst_offset
              djnz      t1, #blnkleer

blnkclr_ret
              ret

LCDdrv_end
              long      0
FIT 496
    